#include "pw/core/vector.hpp"
#include "pw/core/quaternion.hpp"
#include "pw/core/axisangle.hpp"
#include "pw/core/debug.hpp"
#include "pw/core/size.hpp"
#include "pw/core/point.hpp"
#include "pw/core/time.hpp"
#include "pw/core/geometry.hpp"
#include "pw/core/image.hpp"
#include "pw/core/matrix_transform.hpp"
#include "pw/core/rectangle.hpp"

#include "runtime_lua.hpp"


// seems CRTP magic doesnt work with SOL
namespace sol {
template <> struct is_automagical<pw::matrix4x4> : std::false_type {};
template <> struct is_automagical<pw::vector3> : std::false_type {};
template <> struct is_automagical<pw::vector2> : std::false_type {};
template <> struct is_automagical<pw::vector4> : std::false_type {};
template <> struct is_automagical<pw::quaternion> : std::false_type {};
template <> struct is_automagical<pw::rectangle> : std::false_type {};
template <> struct is_automagical<pw::geometry> : std::false_type {};
}

namespace pw {

vector3 table_to_vector3(sol::table t)
{
    return vector3(t[0],t[1],t[2]);
}

template <typename elementType>
std::vector<elementType> convert_sequence(sol::table t)
{
    const std::size_t sz = t.size();
    std::vector<elementType> res(sz);
    for (std::size_t i = 1; i <= sz; i++) {
        res[i - 1] = t[i];
    }
    return res;
}



void register_core_function(sol::state& lua,sol::table& ns)
{

    typedef real_t Scalar;

    ns.set("pi",pw::pi<Scalar>());


    ns.new_usertype<matrix4x4>("matrix4x4"
                               , sol::call_constructor,sol::constructors<matrix4x4()>()
                               , "row",&matrix4x4::row
                               , "column",&matrix4x4::column
                               , "set_identity",&matrix4x4::set_identity
                               , "inverse",&matrix4x4::inverse
                               , "identity",sol::readonly_property(matrix4x4::identity)
                               //                               , "set",[](sol::this_state s,sol::table t) { for (int i = 0;i < t.size();i++) s.Ldata[i] = t[i-1]; })
                               //                               , "get",[]() ->sol::table { return sol::table(); }

                               );

    ns.new_usertype<vector2>("vector2"
                            ,sol::call_constructor,sol::constructors<vector2(),vector2(Scalar,Scalar)>()
                             ,"x", sol::property(sol::resolve<const vector2::value_type&() const>(&vector2::x), [](vector2& v,vector2::value_type val){ v.x() = val;})
            ,"y", sol::property(sol::resolve<const vector2::value_type&() const>(&vector2::y), [](vector2& v,vector2::value_type val){ v.y() = val;})
    );

    ns.new_usertype<vector3>("vector3"
                            ,sol::call_constructor,sol::constructors<vector3(),vector3(Scalar,Scalar,Scalar)>()
                            ,"x", sol::property(sol::resolve<const vector3::value_type&() const>(&vector3::x), [](vector3& v,vector3::value_type val){ v.x() = val;})
                            ,"y", sol::property(sol::resolve<const vector3::value_type&() const>(&vector3::y), [](vector3& v,vector3::value_type val){ v.y() = val;})
                            ,"z", sol::property(sol::resolve<const vector3::value_type&() const>(&vector3::z), [](vector3& v,vector3::value_type val){ v.z() = val;})
                            ,"cross",&vector3::cross
                            ,"transposed",&vector3::transposed
                            ,"lerp",&vector3::lerp
                            ,"forward",&vector3::forward
                            ,"backward",&vector3::backward
                            ,"left",&vector3::left
                            ,"right",&vector3::right
                            ,"up",&vector3::up
                            ,"down",&vector3::down
                            ,sol::meta_function::addition,&vector3::operator+
                            ,sol::meta_function::subtraction,&vector3::operator-
                            ,sol::meta_function::multiplication,&vector3::operator*
                            );

    ns.new_usertype<vector4>("vector4"
                             ,sol::call_constructor,sol::constructors<vector4(),vector4(vector4::value_type,vector4::value_type,vector4::value_type,vector4::value_type)>()
                             ,"x", sol::property(sol::resolve<const vector4::value_type&() const>(&vector4::x), [](vector4& v,vector4::value_type val){ v.x() = val;})
                             ,"y", sol::property(sol::resolve<const vector4::value_type&() const>(&vector4::y), [](vector4& v,vector4::value_type val){ v.y() = val;})
                             ,"z", sol::property(sol::resolve<const vector4::value_type&() const>(&vector4::z), [](vector4& v,vector4::value_type val){ v.z() = val;})
                             ,"w", sol::property(sol::resolve<const vector4::value_type&() const>(&vector4::w), [](vector4& v,vector4::value_type val){ v.w() = val;})
                             ,"project",&vector4::project
                             );


    ns.new_usertype<quaternion>("quaternion"
                                ,sol::constructors<quaternion(), quaternion(quaternion::value_type,quaternion::value_type,quaternion::value_type,quaternion::value_type)>()
                                ,"x", sol::property(sol::resolve<const quaternion::value_type&() const>(&quaternion::x), [](quaternion& v,quaternion::value_type val){ v.x() = val;})
            ,"y", sol::property(sol::resolve<const quaternion::value_type&() const>(&quaternion::y), [](quaternion& v,quaternion::value_type val){ v.y() = val;})
    ,"z", sol::property(sol::resolve<const quaternion::value_type&() const>(&quaternion::z), [](quaternion& v,quaternion::value_type val){ v.z() = val;})
    ,"w", sol::property(sol::resolve<const quaternion::value_type&() const>(&quaternion::w), [](quaternion& v,quaternion::value_type val){ v.w() = val;})
    ,"identity",sol::readonly_property(&quaternion::identity)
            ,"dot",&quaternion::dot
            ,"inverse",sol::readonly_property(&quaternion::inverse)
            ,"normalized",&quaternion::normalized
            ,"lerp",&quaternion::lerp
            ,"slerp",&quaternion::slerp
            ,"matrix",&quaternion::to_matrix
            );

    ns.new_usertype<axisangle>("axisangle",
             sol::constructors<axisangle(), axisangle(vector3,Scalar)>(),
             "axis",&axisangle::axis,
             "angle",&axisangle::angle,
             "from_matrix",&axisangle::from_matrix,
             "to_matrix",&axisangle::to_matrix
             );


    ns.new_usertype<size>("size"
                          , sol::call_constructor, sol::constructors<size(),size(Scalar,Scalar)>()
                          , "width",&size::width
                          , "height",&size::height
                          , "cast_to_float",&size::cast<float>
                          );

    ns.new_usertype<sizef>("sizef"
                          , sol::call_constructor, sol::constructors<sizef(),sizef(Scalar,Scalar)>()
                           , "width",&sizef::width
                           , "height",&sizef::height
                           , "cast_to_int",&sizef::cast<int>
                           );

    ns.new_usertype<point>("point",
                           sol::constructors<point(),point(Scalar,Scalar)>(),
                           "x",&point::x,
                           "y",&point::y
                           );


    ns.new_usertype<debug>("debug",
                           "new",sol::no_constructor,
                           "get",&debug::get,
                           "write",&debug::write
                           //						   "none",sol::debug::level::none
                           );


    ns.new_usertype<time>("time",
                          "now",sol::readonly_property(&time::now),
                          "elapsed",sol::readonly_property(&time::elapsed),
                          "reset",&time::reset
                          );

    ns.new_usertype<geometry>("geometry"
                              ,sol::constructors<geometry(),geometry(geometry::primitive_topology_type,vector3_array,geometry::indices_t)>()
                              ,"primitive_topology", sol::property(&geometry::primitive_topology,&geometry::set_primitive_topology)
                              ,"vertices", sol::property(&geometry::ref_vertices,&geometry::ref_vertices)
                              ,"indices", sol::property(&geometry::ref_indices,&geometry::ref_indices)
//                              ,"texture_coordinates", sol::property(&geometry::ref_texture_coordinates,&geometry::ref_texture_coordinates)
                              ,"compute_normals", &geometry::compute_normals
                              );


    ns.new_enum<false>("primitive_topology_type"
                      ,"point_list", geometry::primitive_topology_type::point_list
                      ,"line_list",  geometry::primitive_topology_type::line_list
                      ,"triangle_list", geometry::primitive_topology_type::triangle_list
                       );


    ns.new_usertype<matrix_transform<real_t>>("matrixtransform"
                                              ,"new",sol::no_constructor
                                              ,"look_at",&matrix_transform<real_t>::look_at
                                              ,"perspective_projection",&matrix_transform<real_t>::perspective_projection);
    // SOL3
//    geoom_type["type"] = lua.create_table_with(
//                  "points", geometry::topology_type::points
//                , "lines", geometry::topology_type::lines
//                , "line_strip", geometry::topology_type::line_strip
//                );



    auto image_type = ns.new_usertype<image>("image"
                           , sol::constructors<image()>()
                                             ,"create",&image::create
                                             ,"size",sol::readonly_property(&image::size)
                                             ,"generate_noise",&image::generate_noise
                                             ,"change_count",sol::property(&image::change_count,&image::set_change_count)
                                             );

    ns.create_named("pixel_layout"
                    ,"rgb8", image::RGB8
                    ,"rgb32", image::RGBA8
                    ,"gray", image::LUM
                    );

    //    image_type["layout"] = lua.create_table_with(
    //                                  "rgb8", image::RGB8
    //                                 ,"rgb32", image::RGBA8
    //                                 ,"gray", image::LUM);


    ns.new_usertype<rectangle>("rectangle"
            ,sol::constructors<rectangle(),rectangle(const point_<float>&,const size_<float>&)>()
                               );

    auto mathf_table = ns.create_named("mathf");
    mathf_table.set_function("ping_pong",ping_pong<real_t>);

}


PW_REGISTER_LUA(core)

}
